use crate::{
    bussy::{channels, Command, Message, PubSub},
    noose::Noose,
    utils::{config::Config, error::Error, structs::Subscription},
};
use nostr_sqlite::{database::NostrDatabase, SQLiteDatabase};
use std::sync::Arc;

pub struct NostrDb {
    db: SQLiteDatabase,
}

impl NostrDb {
    pub async fn new(config: Arc<Config>) -> Self {
        let db_path = config.get_db_path();
        if let Ok(db) = SQLiteDatabase::open(db_path).await {
            return Self { db };
        }

        panic!("[NostrDb] Failed to initialize database");
    }
}

impl Noose for NostrDb {
    async fn start(&mut self, pubsub: Arc<PubSub>) -> Result<(), Error> {
        let mut subscriber = pubsub.subscribe(channels::MSG_NOOSE).await;

        while let Ok(message) = subscriber.recv().await {
            log::info!("[Noose] received message: {:?}", message);
            let command = match message.content {
                Command::DbReqWriteEvent(client_id, event) => match self.write_event(event).await {
                    Ok(status) => Command::DbResOkWithStatus(client_id, status),
                    Err(e) => Command::ServiceError(e),
                },
                Command::DbReqFindEvent(client_id, subscriptioin) => {
                    match self.find_event(subscriptioin).await {
                        Ok(relay_messages) => {
                            Command::DbResRelayMessages(client_id, relay_messages)
                        }
                        Err(e) => Command::ServiceError(e),
                    }
                }
                Command::DbReqEventCounts(client_id, subscriptioin) => {
                    match self.counts(subscriptioin).await {
                        Ok(relay_message) => Command::DbResEventCounts(client_id, relay_message),
                        Err(e) => Command::ServiceError(e),
                    }
                }
                _ => Command::Noop,
            };
            if command != Command::Noop {
                let channel = message.source;
                let message = Message {
                    source: channels::MSG_NOOSE,
                    content: command,
                };

                log::info!(
                    "[Noose] publishing new message: {:?} to channel {}",
                    message,
                    channel
                );

                pubsub.publish(channel, message).await;
            }
        }

        Ok(())
    }

    async fn write_event(&self, event: Box<nostr::Event>) -> Result<nostr::RelayMessage, Error> {
        // TODO: Maybe do event validation and admin deletions here
        match self.db.save_event(&event).await {
            Ok(status) => {
                let relay_message = nostr::RelayMessage::ok(event.id, status, "");
                Ok(relay_message)
            }
            Err(err) => Err(Error::bad_request(err.to_string())),
        }
    }

    async fn find_event(
        &self,
        subscription: Subscription,
    ) -> Result<Vec<nostr::RelayMessage>, Error> {
        match self
            .db
            .query(subscription.filters, nostr_sqlite::database::Order::Desc)
            .await
        {
            Ok(events) => {
                let relay_messages = events
                    .into_iter()
                    .map(|event| nostr::RelayMessage::event(subscription.id.clone(), event))
                    .collect();
                Ok(relay_messages)
            }
            Err(err) => Err(Error::bad_request(err.to_string())),
        }
    }

    async fn counts(&self, subscription: Subscription) -> Result<nostr::RelayMessage, Error> {
        match self.db.count(subscription.filters).await {
            Ok(counts) => {
                let relay_message = nostr::RelayMessage::count(subscription.id, counts);
                Ok(relay_message)
            }
            Err(err) => Err(Error::internal_with_message(err.to_string())),
        }
    }
}
